from abc import abstractmethod

from ..singleton import Singleton
from ..test_case import LLMTestCase
from ..utils import softmax
from sentence_transformers import CrossEncoder


class Metric(metaclass=Singleton):
    # set an arbitrary minimum score that will get over-ridden later
    score: float = 0

    @property
    def minimum_score(self) -> float:
        return self._minimum_score

    @minimum_score.setter
    def minimum_score(self, value: float):
        self._minimum_score = value

    # Measure function signature is subject to be different - not sure
    # how applicable this is - might need a better abstraction
    @abstractmethod
    def measure(self, test_case: LLMTestCase, *args, **kwargs):
        raise NotImplementedError

    def _get_init_values(self):
        # We use this method for sending useful metadata
        init_values = {
            param: getattr(self, param)
            for param in vars(self)
            if isinstance(getattr(self, param), (str, int, float))
        }
        return init_values

    @abstractmethod
    def is_successful(self) -> bool:
        raise NotImplementedError

    @property
    def __name__(self):
        return "Metric"


class EntailmentScoreMetric(Metric):
    def __init__(self, model_name: str = "cross-encoder/nli-deberta-base"):
        # We use a smple cross encoder model

        self.model = CrossEncoder(model_name)

    def measure(self, a: str, b: str):
        scores = self.model.predict([(a, b)])
        # https://huggingface.co/cross-encoder/nli-deberta-base
        # label_mapping = ["contradiction", "entailment", "neutral"]
        return softmax(scores)[0][1]



"""Test BERT score
"""
from deepeval.metrics.bertscore_metric import BertScoreMetric
from deepeval.test_case import LLMTestCase
from deepeval.run_test import assert_test


def test_bert_score():
    metric = BertScoreMetric()
    test_case = LLMTestCase(
        output="Why are you weird", expected_output="Why are you strange?"
    )
    assert_test(test_case, [metric])



# from .test_utils import assert_exact_match, TestEvalCase
# from . import (
#     _version,
#     dataset,
#     evaluator,
#     metrics,
#     query_generator,
#     test_case,
#     test_utils,
#     utils,
# )
import warnings
from ._version import __version__


def check_for_update():
    try:
        import requests

        response = requests.get("https://pypi.org/pypi/deepeval/json")
        latest_version = response.json()["info"]["version"]
        if latest_version > __version__:
            warnings.warn(
                f'You are using deepeval version {__version__}, however version {latest_version} is available. You should consider upgrading via the "pip install --upgrade deepeval" command.'
            )
    except ModuleNotFoundError:
        # they're just getting the version
        pass


check_for_update()



import pytest
import os
from rich import print
from deepeval.api import Api, TestRun
from typing import Optional, Any
from deepeval.constants import PYTEST_RUN_ENV_VAR, PYTEST_RUN_TEST_NAME


def pytest_sessionstart(session: pytest.Session):
    global test_filename
    test_run = TestRun(
        testFile=session.config.getoption("file_or_dir")[0],
        testCases=[],
        metricScores=[],
        configurations={},
    )
    test_filename = test_run.save()


@pytest.hookimpl(tryfirst=True)
def pytest_runtest_protocol(
    item: pytest.Item, nextitem: Optional[pytest.Item]
) -> Optional[Any]:
    os.environ[PYTEST_RUN_TEST_NAME] = item.nodeid.split("::")[-1]
    return None  # continue with the default protocol


@pytest.hookimpl(tryfirst=True, hookwrapper=True)
def pytest_sessionfinish(session: pytest.Session, exitstatus):
    # Code before yield will run before the test teardown

    # yield control back to pytest for the actual teardown
    yield

    # Code after yield will run after the test teardown
    if os.getenv(PYTEST_RUN_ENV_VAR) and os.path.exists(".deepeval"):
        api: Api = Api()
        test_run = TestRun.load(test_filename)
        result = api.post_test_run(test_run)
        link = f"https://app.confident-ai.com/project/{result.projectId}/unit-tests/{result.testRunId}"
        print(
            "✅ Tests finished! View results on " f"[link={link}]{link}[/link]"
        )
        os.remove(test_filename)



"""Tests for answer relevancy
"""
from deepeval.test_case import SearchTestCase
from deepeval.run_test import run_test, assert_test
from deepeval.metrics.ranking_similarity import (
    RankingSimilarity,
    assert_ranking_similarity,
)

list1 = ["a", "b"]
list2 = ["b", "c"]
list3 = ["b", "a"]

list_dict_1 = [{"text": "a"}, {"text": "b"}]
list_dict_2 = [{"text": "b"}, {"text": "c"}]


def test_assert_similar_ranking():
    assert_ranking_similarity(list1, list2, minimum_score=0.4)


def test_similar_ranking_2():
    scorer = RankingSimilarity(minimum_score=0.5)
    test_case = SearchTestCase(list1, list2)
    test_case_2 = SearchTestCase(list1, list3)
    test_results = run_test([test_case, test_case_2], metrics=[scorer])
    assert (
        test_results[1].score > test_results[0].score
    ), "Similarity not working"


def test_similar_ranking_3():
    metric = RankingSimilarity(minimum_score=0.3)
    test_case = SearchTestCase(list1, list2)
    assert_test(test_cases=test_case, metrics=[metric])



from ..singleton import Singleton
from ..test_case import LLMTestCase
from ..run_test import assert_test
from .metric import Metric
import numpy as np


def softmax(x):
    e_x = np.exp(x - np.max(x))
    return e_x / e_x.sum(axis=0)


class AnswerRelevancyModel(metaclass=Singleton):
    def __init__(self):
        from sentence_transformers import SentenceTransformer

        # Load the model
        self.model = SentenceTransformer(
            "sentence-transformers/multi-qa-MiniLM-L6-cos-v1"
        )

    def encode(self, text):
        return self.model.encode(text)


class CrossEncoderAnswerRelevancyModel(metaclass=Singleton):
    def __init__(self, model_name: str = "cross-encoder/nli-deberta-v3-base"):
        from sentence_transformers.cross_encoder import CrossEncoder

        self.model = CrossEncoder(model_name)

    def encode(self, question: str, answer: str):
        scores = self.model.predict([[question, answer]])
        return softmax(scores[0])[2]


class AnswerRelevancyMetric(Metric, metaclass=Singleton):
    def __init__(
        self, minimum_score: bool = 0.5, model_type: str = "cross_encoder"
    ):
        self.minimum_score = minimum_score
        if model_type == "cross_encoder":
            self.model = CrossEncoderAnswerRelevancyModel()
        else:
            self.model = AnswerRelevancyModel()

    def __call__(self, test_case: LLMTestCase):
        score = self.measure(test_case.query, test_case.output)
        self.success = score > self.minimum_score
        return score

    def measure(self, test_case: LLMTestCase) -> float:
        from sentence_transformers import util

        if test_case.query is None or test_case.output is None:
            raise ValueError("query and output cannot be None")

        if isinstance(self.model, CrossEncoderAnswerRelevancyModel):
            score = self.model.encode(test_case.query, test_case.output)
        else:
            docs = [test_case.output]
            # Encode query and documents
            query_emb = self.model.encode(test_case.query)
            doc_emb = self.model.encode(docs)
            # Compute dot score between query and all document embeddings
            scores = util.dot_score(query_emb, doc_emb)[0].cpu().tolist()
            score = scores[0]

        self.success = score > self.minimum_score
        # Log answer relevancy
        self.score = score
        return score

    def is_successful(self) -> bool:
        return self.success

    @property
    def __name__(self):
        return "Answer Relevancy"


def assert_answer_relevancy(
    query: str,
    output: str,
    minimum_score: float = 0.5,
    model_type: str = "default",
):
    metric = AnswerRelevancyMetric(
        minimum_score=minimum_score, model_type=model_type
    )
    test_case = LLMTestCase(query=query, output=output)
    assert_test(test_case, metrics=[metric])



"""Test alert score
"""
import os
import pytest
from deepeval.client import Client
from deepeval.metrics.alert_score import AlertScoreMetric, assert_alert_score
from deepeval.test_case import LLMTestCase
from deepeval.run_test import assert_test

from .utils import assert_viable_score

IMPLEMENTATION_NAME = "Alert"
TEST_API_KEY = "u1s5aFlB6kRyVz/16CZuc7JOQ7e7sCw00N7nfeMZOrk="
os.environ["CONFIDENT_AI_IMP_NAME"] = IMPLEMENTATION_NAME
os.environ["CONFIDENT_AI_API_KEY"] = TEST_API_KEY

query = "Who won the FIFA World Cup in 2018?"
output = "Winners of the FIFA world cup were the French national football team"
expected_output = "French national football team"
context = "The FIFA World Cup in 2018 was won by the French national football team. They defeated Croatia 4-2 in the final match to claim the championship."


def test_alert_score():
    with pytest.raises(AssertionError):
        assert_alert_score(
            query="Who won the FIFA World Cup in 2018?",
            output="Winners of the FIFA world cup were the French national football team",
            expected_output="French national football team",
            context="The FIFA World Cup in 2018 was won by the French national football team. They defeated Croatia 4-2 in the final match to claim the championship.",
        )


@pytest.mark.skip(reason="More research required for Alert Score")
def test_alert_score_metric():
    metric = AlertScoreMetric()
    test_case = LLMTestCase(
        query=query,
        output=output,
        expected_output=expected_output,
        context=context,
    )
    assert_test(test_case, metrics=[metric])


@pytest.mark.skip(reason="More research required for Alert Score")
def test_implementation_inside_overall():
    client = Client(TEST_API_KEY)
    imps = client.list_implementations()
    FOUND = False
    for imp in imps:
        if imp["name"] == IMPLEMENTATION_NAME:
            FOUND = True
    assert (
        FOUND
    ), f"{IMPLEMENTATION_NAME} not found in {[x['name'] for x in imps]}"



"""Simple Retry class
"""
import time
from functools import wraps


class retry:
    retry_count: int = 0
    success_count: int = 0

    def __init__(
        self, max_retries: int = 3, delay: int = 1, min_success: int = 1
    ) -> None:
        self.max_retries = max_retries
        self.delay = delay
        self.min_success = min_success

    def __call__(self, func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            last_error = None  # Store the last error
            while self.retry_count < self.max_retries:
                try:
                    self.retry_count += 1
                    for _ in range(self.min_success - self.success_count):
                        result = func(*args, **kwargs)
                        self.success_count += 1
                except AssertionError as e:
                    print(
                        f"Attempt {self.retry_count} to execute {func.__name__} failed due to AssertionError: {str(e)}"
                    )
                    last_error = e  # Update the last error
                except Exception as e:
                    print(
                        f"Attempt {self.retry_count} to execute {func.__name__} failed due to Exception: {str(e)}"
                    )
                    last_error = e  # Update the last error

                if self.retry_count < self.max_retries:
                    print(
                        f"Retrying execution of {func.__name__} in {self.delay} seconds..."
                    )
                    time.sleep(self.delay)
                else:
                    print(
                        f"Max retries ({self.max_retries}) for executing {func.__name__} exceeded."
                    )
                    if self.success_count < self.min_success:
                        if last_error is not None:
                            raise last_error  # Raise the last error

        return wrapper



import pytest
import typer
import os
import datetime
from typing_extensions import Annotated
from ..metrics.overall_score import assert_overall_score
from .cli_key_handler import set_env_vars
from ..constants import PYTEST_RUN_ENV_VAR

try:
    from rich import print
    from rich.progress import Progress, SpinnerColumn, TextColumn
except Exception as e:
    pass


app = typer.Typer(name="test")


def sample():
    set_env_vars()
    print("Sending sample test results...")
    print(
        "If this is your first time running these models, it may take a while."
    )
    try:
        query = "How does photosynthesis work?"
        output = "Photosynthesis is the process by which green plants and some other organisms use sunlight to synthesize foods with the help of chlorophyll pigment."
        expected_output = "Photosynthesis is the process by which green plants and some other organisms use sunlight to synthesize food with the help of chlorophyll pigment."
        context = "Biology"

        assert_overall_score(query, output, expected_output, context)

    except AssertionError as e:
        pass
    try:
        query = "What is the capital of France?"
        output = "The capital of France is Paris."
        expected_output = "The capital of France is Paris."
        context = "Geography"

        assert_overall_score(query, output, expected_output, context)

    except AssertionError as e:
        pass
    try:
        query = "What are the major components of a cell?"
        output = "Cells have many major components, including the cell membrane, nucleus, mitochondria, and endoplasmic reticulum."
        expected_output = "Cells have several major components, such as the cell membrane, nucleus, mitochondria, and endoplasmic reticulum."
        context = "Biology"
        minimum_score = 0.8  # Adjusting the minimum score threshold

        assert_overall_score(
            query, output, expected_output, context, minimum_score
        )

    except AssertionError as e:
        pass

    try:
        query = "What is the capital of Japan?"
        output = "The largest city in Japan is Tokyo."
        expected_output = "The capital of Japan is Tokyo."
        context = "Geography"

        assert_overall_score(query, output, expected_output, context)
    except AssertionError as e:
        pass

    try:
        query = "Explain the theory of relativity."
        output = "Einstein's theory of relativity is famous."
        expected_output = "Einstein's theory of relativity revolutionized our understanding of space, time, and gravity."
        context = "Physics"

        assert_overall_score(query, output, expected_output, context)
    except AssertionError as e:
        pass


def check_if_legit_file(test_file_or_directory: str):
    if "::" in test_file_or_directory:
        test_file_or_directory, test_case = test_file_or_directory.split("::")
    if os.path.isfile(test_file_or_directory):
        if test_file_or_directory.endswith(".py"):
            if not os.path.basename(test_file_or_directory).startswith("test_"):
                raise ValueError(
                    "Test will not run. Please ensure the file starts with `test_` prefix."
                )
    elif os.path.isdir(test_file_or_directory):
        return
    else:
        raise ValueError(
            "Provided path is neither a valid file nor a directory."
        )


@app.command()
def run(
    test_file_or_directory: str,
    verbose: bool = True,
    color: str = "yes",
    durations: int = 10,
    pdb: bool = False,
    exit_on_first_failure: Annotated[
        bool, typer.Option("--exit-on-first-failure", "-x/-X")
    ] = False,
    show_warnings: Annotated[
        bool, typer.Option("--show-warnings", "-w/-W")
    ] = False,
):
    """Run a test"""
    check_if_legit_file(test_file_or_directory)
    pytest_args = [test_file_or_directory]
    if exit_on_first_failure:
        pytest_args.insert(0, "-x")

    # Generate environment variable based on current date and time
    env_var = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
    os.environ[PYTEST_RUN_ENV_VAR] = env_var

    pytest_args.extend(
        [
            "--verbose" if verbose else "--quiet",
            f"--color={color}",
            f"--durations={durations}",
            "-s",
        ]
    )
    if pdb:
        pytest_args.append("--pdb")
    if not show_warnings:
        pytest_args.append("--disable-warnings")
    # Add the deepeval plugin file to pytest arguments
    pytest_args.extend(["-p", "plugins"])

    retcode = pytest.main(pytest_args)

    # Print this if the run env var is not set
    if not os.getenv(PYTEST_RUN_ENV_VAR):
        print(
            "✅ Tests finished! If logged in, view results on https://app.confident-ai.com/"
        )
    return retcode


@app.command()
def generate(output_file: str = "test_sample.py"):
    with open(output_file, "w") as f:
        f.write(
            """import pytest
from deepeval.metrics.answer_relevancy import AnswerRelevancyMetric
from deepeval.metrics.factual_consistency import FactualConsistencyMetric
from deepeval.metrics.overall_score import assert_overall_score
from deepeval.metrics.randomscore import RandomMetric
from deepeval.test_case import LLMTestCase
from deepeval.run_test import assert_test


def test_1():
    # Check to make sure it is relevant
    query = "What is the capital of France?"
    output = "The capital of France is Paris."
    metric = RandomMetric()
    # Comment this out for differne metrics/models
    # metric = AnswerRelevancyMetric(minimum_score=0.5)
    test_case = LLMTestCase(query=query, output=output)
    assert_test(test_case, [metric])


def test_2():
    # Check to make sure it is factually consistent
    output = "Cells have many major components, including the cell membrane, nucleus, mitochondria, and endoplasmic reticulum."
    context = "Biology"
    metric = RandomMetric()
    # Comment this out for factual consistency tests
    # metric = FactualConsistencyMetric(minimum_score=0.8)
    test_case = LLMTestCase(output=output, context=context)
    assert_test(test_case, [metric])


def test_3():
    # Add a test that fails
    query = "What is the capital of Germany?"
    output = "The capital of Germany is Berlin."
    metric = RandomMetric()
    # metric = AnswerRelevancyMetric(minimum_score=0.5)
    test_case = LLMTestCase(query=query, output=output)
    assert_test(test_case, [metric])


# This is how you can skip a test
@pytest.mark.skip(reason="Can take a while")
def test_0():
    query = "How does photosynthesis work?"
    output = "Photosynthesis is the process by which green plants and some other organisms use sunlight to synthesize foods with the help of chlorophyll pigment."
    expected_output = "Photosynthesis is the process by which green plants and some other organisms use sunlight to synthesize food with the help of chlorophyll pigment."
    context = "Biology"

    assert_overall_score(query, output, expected_output, context)
"""
        )

    print(f"✨ Done! Now run: [bold]deepeval test run {output_file}[/bold]")
    print(
        "You can generate more tests in the future in our documentation at https://docs.confident-ai.com/docs"
    )



from typing import Optional

from ..test_case import LLMTestCase
from .metric import Metric


class CohereRerankerMetric(Metric):
    def __init__(self, api_key: str, minimum_score: float = 0.5):
        try:
            import cohere

            self.cohere_client = cohere.Client(api_key)
        except Exception as e:
            print(e)
            print("Run `pip install cohere`.")
        self.minimum_score = minimum_score

    def __call__(self, output, expected_output, query: Optional[str] = "-"):
        score = self.measure(output, expected_output)
        success = False
        if score > self.minimum_score:
            success = True
        return score

    def measure(self, test_case: LLMTestCase):
        reranked_results = self.cohere_client.rerank(
            query=test_case.query,
            documents=[test_case.output],
            top_n=1,
            model="rerank-english-v2.0",
        )
        score = reranked_results[0].relevance_score
        self.success = score > self.minimum_score
        return score

    def is_successful(self) -> bool:
        return self.success



"""Investigate test case.
"""
import hashlib
from dataclasses import dataclass
from typing import Any, List, Optional, Union


@dataclass
class TestCase:
    id: Optional[str] = None

    def __post_init__(self):
        if self.id is None:
            id_string = "".join(str(value) for value in self.__dict__.values())
            self.id = hashlib.md5(id_string.encode()).hexdigest()


@dataclass
class LLMTestCase(TestCase):
    query: str = "-"
    expected_output: str = "-"
    context: Optional[Union[str, List[str]]] = None
    output: str = "-"

    def __post_init__(self):
        super().__post_init__()
        self.__name__ = f"LLMTestCase_{self.id}"

    # def dict(self):
    #     data = {
    #         "metrics": self.metrics,
    #         "id": self.id,
    #     }
    #     if self.query:
    #         data["query"] = self.query
    #     if self.expected_output:
    #         data["expected_output"] = self.expected_output
    #     if self.context:
    #         data["context"] = self.context
    #     if self.output:
    #         data["output"] = self.output
    #     return data


@dataclass
class SearchTestCase(TestCase):
    def __init__(
        self,
        output_list: List[Any],
        golden_list: List[Any],
        query: Optional[str] = None,
        id: Optional[str] = None,
    ):
        super().__init__(id)
        self.output_list = output_list
        self.golden_list = golden_list
        self.query = query


class AgentTestCase(TestCase):
    """Test Case For Agents"""

    pass



"""Client for Twilix Evals
"""
import getpass
import os
from typing import Dict, List, Optional

from .api import Api
from .get_api_key import _get_api_key


class Client(Api):
    """API with convenience functions"""

    _implementation_id: Dict[str, str] = {}

    def __init__(self, api_key: str = None, local_mode: bool = False, **kwargs):
        if api_key is None or api_key == "":
            api_key = _get_api_key()
        self.api_key = api_key
        self.local_mode = local_mode
        if self.local_mode:
            self.data = []
        super().__init__(api_key=api_key, **kwargs)

    def add_ground_truth(
        self, query: str, expected_response: str, tags: list = None
    ):
        """Add ground truth"""
        if self.local_mode:
            self.data.append(
                {
                    "query": query,
                    "expected_response": expected_response,
                    "tags": tags,
                }
            )
        return self.post_request(
            endpoint="v1/pipeline/eval",
            body={
                "query": query,
                "expected_response": expected_response,
                "tags": tags,
            },
        )

    async def add_ground_truth_async(
        self, query: str, expected_response: str, tags: list = None
    ):
        """Add ground truth"""
        return self.post_request(
            endpoint="v1/pipeline/eval",
            body={
                "query": query,
                "expected_response": expected_response,
                "tags": tags,
            },
        )

    def _get_ground_truths(self, take: int = 50, skip: int = 0):
        return self.get_request(
            endpoint="v1/pipeline/eval", params={"take": take, "skip": skip}
        )

    def _post_evaluation_result(
        self, evaluation_score: float, pipeline_id: str, test_id: str = None
    ):
        return self.post_request(
            endpoint="v1/pipeline/eval/run",
            body={
                "testId": test_id,
                "evaluationScore": evaluation_score,
                "pipelineGroupName": pipeline_id,
            },
        )

    def get_implementation_id_by_name(
        self, name: str, description: Optional[str] = None
    ):
        """Gets implementation. If none exists, it creates one."""
        if name in self._implementation_id:
            return self._implementation_id[name]
        implementations: List[Dict] = self.list_implementations()
        imp_id = None
        for imp in implementations:
            if imp["name"] == name:
                imp_id = imp["id"]
        if imp_id is None:
            created_imp = self.create_implementation(
                name=name, description=description
            )
            imp_id = created_imp["id"]
        # to avoid hammering the server
        self._implementation_id[name] = imp_id
        return imp_id



# Sample Metric for BLEU
import numpy as np

from .metric import Metric


class BLEUMetric(Metric):
    def __init__(self, minimum_score: float = 0.5):
        self.minimum_score = minimum_score

    def compute_bleu(
        self, candidate: str, reference: str, weights=(0.25, 0.25, 0.25, 0.25)
    ):
        """
        Compute BLEU score for a candidate sentence given a reference sentence.

        :param candidate: The candidate sentence as a string.
        :param reference: The reference sentence as a string.
        :param weights: Weights for the n-gram precisions, default is uniform (0.25 for each).
        :return: BLEU score.
        """
        import nltk
        from nltk.util import ngrams

        candidate = (
            candidate.split()
        )  # Convert the candidate string to a list of tokens
        reference = (
            reference.split()
        )  # Convert the reference string to a list of tokens

        precisions = []

        for i in range(1, 5):  # Compute BLEU for 1 to 4-grams
            candidate_ngrams = ngrams(candidate, i)
            candidate_ngram_freq = nltk.FreqDist(candidate_ngrams)

            reference_ngrams = ngrams(reference, i)
            reference_ngram_freq = nltk.FreqDist(reference_ngrams)

            clipped_counts = {
                ngram: min(
                    candidate_ngram_freq[ngram], reference_ngram_freq[ngram]
                )
                for ngram in candidate_ngram_freq
            }
            precision = sum(clipped_counts.values()) / sum(
                candidate_ngram_freq.values()
            )
            precisions.append(precision)

        brevity_penalty = min(1, len(candidate) / len(reference))

        bleu = brevity_penalty * np.exp(
            np.mean([w * np.log(p) for w, p in zip(weights, precisions)])
        )

        return bleu



"""Tests for answer length
"""
from deepeval.test_case import LLMTestCase
from deepeval.metrics.answer_length import LengthMetric
from deepeval.run_test import assert_test


def test_answer_length():
    metric = LengthMetric()
    test_case = LLMTestCase(
        output=" Some output ", expected_output="Some output"
    )
    assert_test(test_case, [metric])



"""Pipeline
"""
from typing import Callable

from .metrics.metric import Metric


class Pipeline:
    def __init__(self, pipeline_id: str, result_function: Callable):
        self.pipeilne_id = pipeline_id
        self.result_function = result_function



"""Alert Score
"""
from ..test_case import LLMTestCase
from .entailment_metric import EntailmentScoreMetric
from .metric import Metric
from ..run_test import assert_test


class AlertScoreMetric(Metric):
    def __init__(self, minimum_score: float = 0.5):
        self.minimum_score = minimum_score
        self.entailment_metric = EntailmentScoreMetric()
        # self.answer_relevancy = AnswerRelevancyMetric()

    def __call__(self, test_case: LLMTestCase):
        score = self.measure(test_case)
        return score

    def measure(self, test_case: LLMTestCase) -> float:
        entailment_score = self.entailment_metric.measure(test_case)

        answer_expected_score = self.entailment_metric.measure(test_case)

        # This metric is very very bad right now as it requires the answer
        # to re-gurgitate the question.
        # answer_relevancy_score = self.answer_relevancy.measure(
        #     query=query, answer=output
        # )
        alert_score = min(entailment_score, answer_expected_score)
        self.success = bool(alert_score > self.minimum_score)
        return alert_score

    def is_successful(self) -> bool:
        return self.success

    @property
    def __name__(self):
        return "Alert Score"


def assert_alert_score(
    query: str,
    output: str,
    expected_output: str,
    context: str,
    minimum_score: float = 0.5,
):
    """Create alert score."""
    metric = AlertScoreMetric(minimum_score=minimum_score)
    test_case = LLMTestCase(
        query=query,
        expected_output=expected_output,
        context=context,
        output=output,
    )
    assert_test(test_case, [metric])



"""Tests for toxic
"""


import pytest
from deepeval.test_case import LLMTestCase
from deepeval.metrics.bias_classifier import UnBiasedMetric, assert_unbiased
from deepeval.run_test import assert_test

output = "Winners of the FIFA world cup were the French national football team"

metric = UnBiasedMetric()


def test_non_toxic():
    assert_test(
        test_cases=LLMTestCase(
            output=output,
        ),
        metrics=[metric],
    )


def test_non_toxic_2():
    with pytest.raises(AssertionError):
        assert_unbiased(
            text="Left wing is evil.",
        )


def test_non_toxic_metric():
    metric = UnBiasedMetric()
    test_case = LLMTestCase(output=output)
    assert_test(test_case, [metric])
    assert_test(test_case, [metric])


def test_bias_score():
    metric = UnBiasedMetric()
    test_case = LLMTestCase(output=output)
    assert_test(test_case, [metric])



"""Metric for toxic classifier. 
1 - Healthy
0 - Toxic
"""
from ..singleton import Singleton
from ..test_case import LLMTestCase
from .metric import Metric
from ..run_test import assert_test


class DetoxifyModel(metaclass=Singleton):
    def __init__(self, model_name: str = "original"):
        self.model_name = model_name

        try:
            from detoxify import Detoxify
        except ImportError as e:
            print(e)
        self.model = Detoxify(model_name)

    def predict(self, text: str):
        return self.model.predict(text)


class NonToxicMetric(Metric):
    def __init__(
        self, model_name: str = "original", minimum_score: float = 0.5
    ):
        self.detoxify_model = DetoxifyModel(model_name)
        self.minimum_score = minimum_score

    def __call__(self, test_case: LLMTestCase):
        score = self.measure(test_case.output)
        score = score["min_score"]
        return score

    def measure(self, test_case: LLMTestCase):
        if test_case.output is None:
            raise ValueError("output cannot be None")
        results = self.detoxify_model.predict(test_case.output)
        # sample output
        # {'toxicity': 0.98057544,
        # 'severe_toxicity': 0.106649496,
        # 'obscene': 0.32923067,
        # 'threat': 0.018646086,
        # 'insult': 0.7514423,
        # 'identity_attack': 0.86643445}
        self.success = True
        for i, (k, v) in enumerate(results.items()):
            if i == 0:
                min_score = 1 - v
            if v > 1 - self.minimum_score:
                self.success = False
                min_score = 1 - v
        results["min_score"] = min_score
        print(results)
        return min_score

    def is_successful(self):
        return self.success

    @property
    def __name__(self):
        return "Toxicness'"


def assert_non_toxic(text: str, minimum_score: float = 0.5):
    metric = NonToxicMetric(minimum_score=minimum_score)
    test_case = LLMTestCase(output=text)
    assert_test(test_case, [metric])



from .metric import Metric
from deepeval.test_case import LLMTestCase


class LengthMetric(Metric):
    """This metric returns the length of the output if the output exists and removes any spaces on each end"""

    def __init__(self, minimum_length: int = 0):
        self.minimum_length = minimum_length

    def measure(self, test_case: LLMTestCase):
        # sends to server
        score = len(test_case.output.strip())
        self.success = score > self.minimum_length
        return score

    def is_successful(self):
        return self.success

    @property
    def __name__(self):
        return "AnswerLength"



"""Tests for answer relevancy
"""
import pytest
from deepeval.test_case import LLMTestCase
from deepeval.metrics.answer_relevancy import (
    AnswerRelevancyMetric,
    assert_answer_relevancy,
)
from deepeval.run_test import run_test, assert_test
from .utils import assert_viable_score

query = "What is Python?"
answer = "Python is a programming language?"


def test_answer_relevancy():
    assert_answer_relevancy(query, answer, minimum_score=0.5)


def test_answer_not_relevant():
    with pytest.raises(AssertionError):
        assert_answer_relevancy(query, "He is not your friend")


def test_query_answer_relevancy():
    scorer = AnswerRelevancyMetric(minimum_score=0.5)
    test_case = LLMTestCase(query=query, output=answer)
    assert_test(test_case, [scorer])


def test_compare_answer_relevancy_2():
    scorer = AnswerRelevancyMetric(minimum_score=0.5)
    test_case = LLMTestCase(query=query, output="Programming lang")
    test_case_2 = LLMTestCase(
        query=query, output="Python is a programming lang"
    )
    results = run_test([test_case, test_case_2], metrics=[scorer])
    assert results[1].score > results[0].score


def test_compare_answer_relevancy():
    metric = AnswerRelevancyMetric(minimum_score=0.5)
    query = "what is python"
    test_case = LLMTestCase(query=query, output="Programming lang")
    test_case_2 = LLMTestCase(
        query=query, output="Python is a programming lang"
    )
    result = run_test([test_case, test_case_2], metrics=[metric])
    assert result[1].score > result[0].score


def test_cross_encoder_answer_relevancy():
    scorer = AnswerRelevancyMetric(
        minimum_score=0.5, model_type="cross_encoder"
    )
    test_case = LLMTestCase(query=query, output=answer)
    score = assert_test(test_case, [scorer])
    assert_viable_score(score[0].score)



"""Class for Evaluation Datasets
"""
import json
import random
import time
from collections import UserList
from datetime import datetime
from typing import Any, Callable, List, Optional

from tabulate import tabulate

from deepeval.run_test import run_test
from deepeval.metrics.metric import Metric
from deepeval.test_case import LLMTestCase
from dataclasses import asdict


class EvaluationDataset(UserList):
    """Class for Evaluation Datasets -  which are a list of test cases"""

    def __init__(self, test_cases: List[LLMTestCase]):
        self.data: List[LLMTestCase] = test_cases

    @classmethod
    def from_csv(
        cls,  # Use 'cls' instead of 'self' for class methods
        csv_filename: str,
        query_column: Optional[str] = None,
        expected_output_column: Optional[str] = None,
        context_column: Optional[str] = None,
        output_column: Optional[str] = None,
        id_column: str = None,
        metrics: List[Metric] = None,
    ):
        import pandas as pd

        df = pd.read_csv(csv_filename)
        if query_column is not None and query_column in df.columns:
            querys = df[query_column].values
        else:
            querys = [None] * len(df)
        if (
            expected_output_column is not None
            and expected_output_column in df.columns
        ):
            expected_outputs = df[expected_output_column].values
        else:
            expected_outputs = [None] * len(df)
        if context_column is not None and context_column in df.columns:
            contexts = df[context_column].values
        else:
            contexts = [None] * len(df)
        if output_column is not None and output_column in df.columns:
            outputs = df[output_column].values
        else:
            outputs = [None] * len(df)
        if id_column is not None:
            ids = df[id_column].values
        else:
            ids = [None] * len(df)

        # Initialize the 'data' attribute as an empty list
        cls.data = []

        for i, query_data in enumerate(querys):
            cls.data.append(
                LLMTestCase(
                    query=query_data,
                    expected_output=expected_outputs[i],
                    context=contexts[i],
                    id=ids[i] if id_column else None,
                    output=outputs[i] if output_column else None,
                )
            )
        return cls(cls.data)

    def from_test_cases(self, test_cases: list):
        self.data = test_cases

    @classmethod
    def from_hf_dataset(
        cls,
        dataset_name: str,
        split: str,
        query_column: str,
        expected_output_column: str,
        context_column: str = None,
        output_column: str = None,
        id_column: str = None,
    ):
        """
        Load test cases from a HuggingFace dataset.

        Args:
            dataset_name (str): The name of the HuggingFace dataset to load.
            split (str): The split of the dataset to load (e.g., 'train', 'test').
            query_column (str): The column in the dataset corresponding to the query.
            expected_output_column (str): The column in the dataset corresponding to the expected output.
            context_column (str, optional): The column in the dataset corresponding to the context. Defaults to None.
            output_column (str, optional): The column in the dataset corresponding to the output. Defaults to None.
            id_column (str, optional): The column in the dataset corresponding to the ID. Defaults to None.

        Returns:
            EvaluationDataset: An instance of EvaluationDataset containing the loaded test cases.
        """
        try:
            from datasets import load_dataset
        except ImportError:
            raise ImportError(
                "The 'datasets' library is missing. Please install it using pip: pip install datasets"
            )

        hf_dataset = load_dataset(dataset_name, split=split)
        test_cases = []

        for i, row in enumerate(hf_dataset):
            test_cases.append(
                LLMTestCase(
                    query=row[query_column],
                    expected_output=row[expected_output_column],
                    context=row[context_column] if context_column else None,
                    output=row[output_column] if output_column else None,
                    id=row[id_column] if id_column else None,
                )
            )
        return cls(test_cases)

    @classmethod
    def from_json(
        cls,
        json_filename: str,
        query_column: str,
        expected_output_column: str,
        context_column: str,
        output_column: str,
        id_column: str = None,
    ):
        """
        This is for JSON data in the format of key-value array pairs.
        {
            "query": ["What is the customer success number", "What is the customer success number"],
            "context": ["Context 1", "Context 2"],
            "output": ["Output 1", "Output 2"]
        }

        if the JSON data is in a list of dictionaries, use from_json_list
        """
        with open(json_filename, "r") as f:
            data = json.load(f)
        test_cases = []

        for i, query in enumerate(data[query_column]):
            test_cases.append(
                LLMTestCase(
                    query=data[query_column][i],
                    expected_output=data[expected_output_column][i],
                    context=data[context_column][i],
                    output=data[output_column][i],
                    id=data[id_column][i] if id_column else None,
                )
            )
        return cls(test_cases)

    @classmethod
    def from_json_list(
        cls,
        json_filename: str,
        query_column: str,
        expected_output_column: str,
        context_column: str,
        output_column: str,
        id_column: str = None,
    ):
        """
        This is for JSON data in the format of a list of dictionaries.
        [
            {"query": "What is the customer success number", "expected_output": "What is the customer success number", "context": "Context 1", "output": "Output 1"},
        ]
        """
        with open(json_filename, "r") as f:
            data = json.load(f)
        test_cases = []
        for i, query in enumerate(data):
            test_cases.append(
                LLMTestCase(
                    query=data[i][query_column],
                    expected_output=data[i][expected_output_column],
                    context=data[i][context_column],
                    output=data[i][output_column],
                    id=data[i][id_column] if id_column else None,
                )
            )
        return cls(test_cases)

    @classmethod
    def from_dict(
        cls,
        data: List[dict],
        query_key: str,
        expected_output_key: str,
        context_key: str = None,
        output_key: str = None,
        id_key: str = None,
    ):
        """
        Load test cases from a list of dictionaries.

        Args:
            data (List[dict]): The list of dictionaries containing the test case data.
            query_key (str): The key in each dictionary corresponding to the query.
            expected_output_key (str): The key in each dictionary corresponding to the expected output.
            context_key (str, optional): The key in each dictionary corresponding to the context. Defaults to None.
            output_key (str, optional): The key in each dictionary corresponding to the output. Defaults to None.
            id_key (str, optional): The key in each dictionary corresponding to the ID. Defaults to None.
            metrics (List[Metric], optional): The list of metrics to be associated with the test cases. Defaults to None.

        Returns:
            EvaluationDataset: An instance of EvaluationDataset containing the loaded test cases.
        """
        test_cases = []
        for i, case_data in enumerate(data):
            test_cases.append(
                LLMTestCase(
                    query=case_data[query_key],
                    expected_output=case_data[expected_output_key],
                    context=case_data[context_key] if context_key else None,
                    output=case_data[output_key] if output_key else None,
                    id=case_data[id_key] if id_key else None,
                )
            )
        return cls(test_cases)

    def to_dict(self):
        return [asdict(x) for x in self.data]

    def to_csv(self, csv_filename: str):
        import pandas as pd

        df = pd.DataFrame(self.data)
        df.to_csv(csv_filename, index=False)

    def to_json(self, json_filename: str):
        with open(json_filename, "w") as f:
            json.dump(self.data, f)

    def from_hf_evals(self):
        raise NotImplementedError

    def from_df(self):
        raise NotImplementedError

    def __repr__(self):
        return f"{self.__class__.__name__}({self.data})"

    def sample(self, n: int = 5):
        if len(self.data) <= n:
            n = len(self.data)
        result = random.sample(self.data, n)
        return [asdict(r) for r in result]

    def head(self, n: int = 5):
        return self.data[:n]

    def __getitem__(self, index):
        return self.data[index]

    def __setitem__(self, index, value):
        self.data[index] = value

    def __delitem__(self, index):
        del self.data[index]

    def run_evaluation(
        self,
        completion_fn: Callable[[str], str] = None,
        outputs: List[str] = None,
        test_filename: str = None,
        max_retries: int = 3,
        min_success: int = 1,
        metrics: List[Metric] = None,
    ) -> str:
        """Run evaluation with given metrics"""
        if completion_fn is None:
            assert outputs is not None

        table: List[List[Any]] = []

        headers: List[str] = [
            "Test Passed",
            "Metric Name",
            "Score",
            "Output",
            "Expected output",
            "Message",
        ]
        results = run_test(
            test_cases=self.data,
            metrics=metrics,
            raise_error=True,
            max_retries=max_retries,
            min_success=min_success,
        )
        for result in results:
            table.append(
                [
                    result.success,
                    result.metric_name,
                    result.score,
                    result.output,
                    result.expected_output,
                    "",
                ]
            )
        if test_filename is None:
            test_filename = (
                f"test-result-{datetime.now().__str__().replace(' ', '-')}.txt"
            )
        with open(test_filename, "w") as f:
            f.write(tabulate(table, headers=headers))
        print(f"Saved to {test_filename}")
        for t in table:
            assert t[0] == True, t[-1]
        return test_filename

    def review(self):
        """A bulk editor for reviewing synthetic data."""
        try:
            from dash import (
                Dash,
                Input,
                Output,
                State,
                callback,
                dash_table,
                dcc,
                html,
            )
        except ModuleNotFoundError:
            raise Exception(
                """You will need to run `pip install dash` to be able to review tests that were automatically created."""
            )

        table_data = [
            {"query": x.query, "expected_output": x.expected_output}
            for x in self.data
        ]
        app = Dash(
            __name__,
            external_stylesheets=[
                "https://cdn.jsdelivr.net/npm/bootswatch@5.3.1/dist/darkly/bootstrap.min.css"
            ],
        )

        app.layout = html.Div(
            [
                html.H1("Bulk Review Test Cases", style={"marginLeft": "20px"}),
                html.Button(
                    "Add Test case",
                    id="editing-rows-button",
                    n_clicks=0,
                    style={
                        "padding": "8px",
                        "backgroundColor": "purple",  # Added purple background color
                        "color": "white",
                        "border": "2px solid purple",  # Added purple border
                        "marginLeft": "20px",
                    },
                ),
                html.Div(
                    dash_table.DataTable(
                        id="adding-rows-table",
                        columns=[
                            {
                                "name": c.title().replace("_", " "),
                                "id": c,
                                "deletable": True,
                                "renamable": True,
                            }
                            for i, c in enumerate(["query", "expected_output"])
                        ],
                        data=table_data,
                        editable=True,
                        row_deletable=True,
                        style_data_conditional=[
                            {
                                "if": {"row_index": "odd"},
                                "backgroundColor": "rgb(40, 40, 40)",
                                "color": "white",
                            },
                            {
                                "if": {"row_index": "even"},
                                "backgroundColor": "rgb(30, 30, 30)",
                                "color": "white",
                            },
                            {
                                "if": {"state": "selected"},
                                "backgroundColor": "white",
                                "color": "white",
                            },
                        ],
                        style_header={
                            "backgroundColor": "rgb(30, 30, 30)",
                            "color": "white",
                            "fontWeight": "bold",
                            "padding": "10px",  # Added padding
                        },
                        style_cell={
                            "padding": "10px",  # Added padding
                            "whiteSpace": "pre-wrap",  # Wrap cell contents
                            "maxHeight": "200px",
                        },
                    ),
                    style={"padding": "20px"},  # Added padding
                ),
                html.Div(style={"margin-top": "20px"}),
                html.Button(
                    "Save To CSV",
                    id="save-button",
                    n_clicks=0,
                    style={
                        "padding": "8px",
                        "backgroundColor": "purple",  # Added purple background color
                        "color": "white",
                        "border": "2px solid purple",  # Added purple border
                        "marginLeft": "20px",
                    },
                ),
                dcc.Input(
                    id="filename-input",
                    type="text",
                    placeholder="Enter filename (.csv format)",
                    style={
                        "padding": "8px",
                        "backgroundColor": "rgb(30, 30, 30)",
                        "color": "white",
                        "marginLeft": "20px",
                        "border": "2px solid purple",  # Added purple border
                        "width": "200px",  # Edited width
                    },
                    value="review-test.csv",
                ),
                html.Div(id="code-output"),
            ],
            style={"padding": "20px"},  # Added padding
        )

        @callback(
            Output("adding-rows-table", "data"),
            Input("editing-rows-button", "n_clicks"),
            State("adding-rows-table", "data"),
            State("adding-rows-table", "columns"),
        )
        def add_row(n_clicks, rows, columns):
            if n_clicks > 0:
                rows.append({c["id"]: "" for c in columns})
            return rows

        @callback(
            Output("save-button", "n_clicks"),
            Input("save-button", "n_clicks"),
            State("adding-rows-table", "data"),
            State("adding-rows-table", "columns"),
            State("filename-input", "value"),
        )
        def save_data(n_clicks, rows, columns, filename):
            if n_clicks > 0:
                import csv

                with open(filename, "w", newline="") as f:
                    writer = csv.DictWriter(
                        f, fieldnames=[c["id"] for c in columns]
                    )
                    writer.writeheader()
                    writer.writerows(rows)
            return n_clicks

        @app.callback(
            Output("code-output", "children"),
            Input("save-button", "n_clicks"),
            State("filename-input", "value"),
        )
        def show_code(n_clicks, filename):
            if n_clicks > 0:
                code = f"""
        from deepeval.dataset import EvaluationDataset

        # Replace 'filename.csv' with the actual filename
        ds = EvaluationDataset.from_csv('{filename}')

        # Access the data in the CSV file
        # For example, you can print a few rows
        print(ds.sample())
        """
                return html.Div(
                    [
                        html.P(
                            "Code to load the CSV file back into a dataset for testing:"
                        ),
                        html.Pre(code, className="language-python"),
                    ],
                    style={"padding": "20px"},  # Added padding
                )
            else:
                return ""

        app.run(debug=False)

    def add_evaluation_query_answer_pairs(
        self,
        openai_api_key: str,
        context: str,
        n: int = 3,
        model: str = "openai/gpt-3.5-turbo",
    ):
        """Utility function to create an evaluation dataset using ChatGPT."""
        new_dataset = create_evaluation_query_answer_pairs(
            openai_api_key=openai_api_key, context=context, n=n, model=model
        )
        self.data += new_dataset.data
        print(f"Added {len(new_dataset.data)}!")


def make_chat_completion_request(prompt: str, openai_api_key: str):
    import openai

    openai.api_key = openai_api_key
    response = openai.ChatCompletion.create(
        model="gpt-3.5-turbo",
        messages=[
            {"role": "system", "content": "You are a helpful assistant."},
            {"role": "user", "content": prompt},
        ],
    )
    return response.choices[0].message.content


def generate_chatgpt_output(prompt: str, openai_api_key: str) -> str:
    max_retries = 3
    retry_delay = 1
    for attempt in range(max_retries):
        try:
            expected_output = make_chat_completion_request(
                prompt=prompt, openai_api_key=openai_api_key
            )
            break
        except Exception as e:
            print(f"Error occurred: {e}")
            if attempt < max_retries - 1:
                print(f"Retrying in {retry_delay} seconds...")
                time.sleep(retry_delay)
                retry_delay *= 2
            else:
                raise

    return expected_output


def create_evaluation_query_answer_pairs(
    openai_api_key: str,
    context: str,
    n: int = 3,
    model: str = "openai/gpt-3.5-turbo",
) -> EvaluationDataset:
    """Utility function to create an evaluation dataset using GPT."""
    prompt = f"""You are generating {n} sets of of query-answer pairs to create an evaluation dataset based on the below context.
Context: {context}

Respond in JSON format in 1 single line without white spaces an array of JSON with the keys `query` and `answer`.
"""
    for _ in range(3):
        try:
            responses = generate_chatgpt_output(
                prompt, openai_api_key=openai_api_key
            )
            responses = json.loads(responses)
            break
        except Exception as e:
            return EvaluationDataset(test_cases=[])

    test_cases = []
    for response in responses:
        test_case = LLMTestCase(
            query=response["query"],
            expected_output=response["answer"],
            context=context,
        )
        test_cases.append(test_case)

    dataset = EvaluationDataset(test_cases=test_cases)
    return dataset



"""Function for running test
"""
import os
from typing import List, Optional, Union
import time
from dataclasses import dataclass
from .retry import retry
from .client import Client
from .constants import (
    IMPLEMENTATION_ID_ENV,
    LOG_TO_SERVER_ENV,
    PYTEST_RUN_ENV_VAR,
)
from .get_api_key import _get_api_key, _get_implementation_name
from .metrics import Metric
from .test_case import LLMTestCase, TestCase, SearchTestCase
from .api import TestRun


def _is_api_key_set():
    result = _get_api_key()
    # if result == "" or result is None:
    #     warnings.warn(
    #         """API key is not set - you won't be able to log to the DeepEval dashboard. Please set it by running `deepeval login`"""
    #     )
    if result == "" or result is None:
        return False
    return True


def _is_send_okay():
    # DOing this until the API endpoint is fixed
    return _is_api_key_set() and os.getenv(LOG_TO_SERVER_ENV) != "Y"


def _get_init_values(metric: Metric):
    # We use this method for sending useful metadata
    init_values = {
        param: getattr(metric, param)
        for param in vars(metric)
        if isinstance(getattr(metric, param), (str, int, float))
    }
    return init_values


@dataclass
class TestResult:
    """Returned from run_test"""

    success: bool
    score: float
    metric_name: str
    query: str
    output: str
    expected_output: str
    metadata: Optional[dict]
    context: str

    def __gt__(self, other: "TestResult") -> bool:
        """Greater than comparison based on score"""
        return self.score > other.score

    def __lt__(self, other: "TestResult") -> bool:
        """Less than comparison based on score"""
        return self.score < other.score


def create_test_result(
    test_case: Union[LLMTestCase, SearchTestCase],
    success: bool,
    score: float,
    metric: float,
) -> TestResult:
    if isinstance(test_case, LLMTestCase):
        return TestResult(
            success=success,
            score=score,
            metric_name=metric.__name__,
            query=test_case.query if test_case.query else "-",
            output=test_case.output if test_case.output else "-",
            expected_output=test_case.expected_output
            if test_case.expected_output
            else "-",
            metadata=None,
            context=test_case.context,
        )
    elif isinstance(test_case, SearchTestCase):
        return TestResult(
            success=success,
            score=score,
            metric_name=metric.__name__,
            query=test_case.query if test_case.query else "-",
            output=test_case.output_list if test_case.output_list else "-",
            expected_output=test_case.golden_list
            if test_case.golden_list
            else "-",
            metadata=None,
            context="-",
        )
    else:
        raise ValueError("TestCase not supported yet.")


def run_test(
    test_cases: Union[TestCase, LLMTestCase, SearchTestCase, List[LLMTestCase]],
    metrics: List[Metric],
    max_retries: int = 1,
    delay: int = 1,
    min_success: int = 1,
    raise_error: bool = False,
) -> List[TestResult]:
    """
    Args:
        test_cases: Either a single test case or a list of test cases to run
        metrics: List of metrics to run
        raise_error: Whether to raise an error if a metric fails
        max_retries: Maximum number of retries for each metric measurement
        delay: Delay in seconds between retries
        min_success: Minimum number of successful measurements required

    Example:
        >>> from deepeval.metrics.facutal_consistency import FactualConsistencyMetric
        >>> from deepeval.test_case import LLMTestCase
        >>> from deepeval.run_test import run_test
        >>> metric = FactualConsistencyMetric()
        >>> test_case = LLMTestCase(
        ...     query="What is the capital of France?",
        ...     output="Paris",
        ...     expected_output="Paris",
        ...     context="Geography",
        ... )
        >>> run_test(test_case, metric)
    """
    if isinstance(test_cases, TestCase):
        test_cases = [test_cases]

        test_results = []
        for test_case in test_cases:
            failed_metrics = []
            for metric in metrics:
                test_start_time = time.perf_counter()

                @retry(
                    max_retries=max_retries,
                    delay=delay,
                    min_success=min_success,
                )
                def measure_metric():
                    score = metric.measure(test_case)
                    success = metric.is_successful()
                    test_result = create_test_result(
                        test_case, success, score, metric
                    )
                    test_results.append(test_result)

                    # Load the test_run and add the test_case regardless of the success of the test
                    test_end_time = time.perf_counter()
                    run_duration = test_end_time - test_start_time
                    if os.getenv(PYTEST_RUN_ENV_VAR):
                        test_run = TestRun.load()
                        metric.score = score
                        test_run.add_llm_test_case(
                            test_case=test_case,
                            metrics=[metric],
                            run_duration=run_duration,
                        )
                        test_run.save()

                    if not success:
                        failed_metrics.append((metric.__name__, score))

                measure_metric()

        if raise_error and failed_metrics:
            raise AssertionError(
                f"Metrics {', '.join([f'{name} (Score: {score})' for name, score in failed_metrics])} failed."
            )

    return test_results


def assert_test(
    test_cases: Union[LLMTestCase, List[LLMTestCase]],
    metrics: List[Metric],
    max_retries: int = 1,
    delay: int = 1,
    min_success: int = 1,
) -> List[TestResult]:
    """Assert a test"""
    return run_test(
        test_cases=test_cases,
        metrics=metrics,
        max_retries=max_retries,
        delay=delay,
        min_success=min_success,
        raise_error=True,
    )



import os
from typing import Optional

import typer
from typing_extensions import Annotated

# Rich has a few dependency issues
try:
    from rich import print
except Exception as e:
    pass
from ..api import Api
from ..constants import IMPLEMENTATION_ID_NAME
from ..key_handler import KEY_FILE_HANDLER
from .test import app as test_app

app = typer.Typer(name="deepeval")

app.add_typer(test_app, name="test")


@app.command()
def login(
    api_key: Annotated[
        str,
        typer.Option(
            help="API key to get from https://app.confident-ai.com. Required if you want to log events to the server."
        ),
    ] = "",
    implementation_name: Annotated[
        str, typer.Option(help="The name of your project")
    ] = "",
):
    """Login to the DeepEval platform."""
    print("Welcome to [bold]DeepEval[/bold]!")
    print(
        "Grab your API key here: [link=https://app.confident-ai.com]https://app.confident-ai.com[/link] "
    )
    if api_key == "":
        api_key = KEY_FILE_HANDLER.fetch_api_key()
        api_key = typer.prompt(
            text="Paste it here (Hit enter if default is right)",
            default=api_key,
        )
    KEY_FILE_HANDLER.write_api_key(api_key)
    client = Api(api_key=api_key)
    if implementation_name == "":
        implementation_name = KEY_FILE_HANDLER.fetch_implementation_name()
        if implementation_name is None or implementation_name == "":
            implementation_name = "example"
        print("What is the name of your project?")
        implementation_name = typer.prompt(
            text="Name (Hit enter if default is right):",
            default=implementation_name,
        )
    KEY_FILE_HANDLER.write_data(IMPLEMENTATION_ID_NAME, implementation_name)
    print("Success! :raising_hands:")
    print(
        "If you are new to DeepEval, try generate a sample test: [bold]deepeval test generate --output-file test_sample.py[/bold]"
    )
    print("Run a sample test: [bold]deepeval test run test_sample.py[/bold]")
    print(
        "To switch to a different project: [bold]deepeval switch --implementation_name your_project_name[/bold]"
    )


@app.command()
def switch(
    implementation_name: Annotated[
        str, typer.Option(help="The name of the project you want to switch to")
    ] = "",
):
    """Switch to a different project on the DeepEval platform."""
    if implementation_name == "":
        print("You must provide a project name to switch to.")
    else:
        KEY_FILE_HANDLER.write_data(IMPLEMENTATION_ID_NAME, implementation_name)
        print(f"Switched to project: {implementation_name}")


if __name__ == "__main__":
    app()



"""Metric for bias classifier - using the same min score subtraction methodology as the toxic classifier 
Rationale for bias classifier is described here https://arxiv.org/pdf/2208.05777.pdf
1 - Not Biased
0 - Bias
"""

import warnings
from typing import Optional
from .metric import Metric

from ..singleton import Singleton
from ..test_case import LLMTestCase
from ..run_test import assert_test


class UnBiasedMetric(Metric, metaclass=Singleton):
    def __init__(
        self, model_name: str = "original", minimum_score: float = 0.5
    ):  # see paper for rationale https://arxiv.org/pdf/2208.05777.pdf
        self.model_name = model_name
        self.minimum_score = minimum_score

    def __call__(self, output, expected_output, query: Optional[str] = "-"):
        score = self.measure(output, expected_output)
        success = score >= self.minimum_score
        return score

    def measure(self, test_case: LLMTestCase, return_all_scores: bool = False):
        if test_case.output is None:
            raise ValueError("Required attributes for test_case cannot be None")

        from Dbias.bias_classification import classifier

        warnings.warn(
            "Run `pip install deepeval[bias]`. If you have, please ignore this warning."
        )

        results = classifier(test_case.output)
        # sample output
        # [{'label': 'Biased', 'score': 0.9938021898269653}]
        if results[0]["label"] == "Biased":
            v = 0.5 - (results[0]["score"] / 2)
        else:
            # if it's unbiased - use normal score
            v = 0.5 + (results[0]["score"] / 2)

        self.success = False
        if v > self.minimum_score:
            self.success = True

        self.score = v
        if return_all_scores:
            return results
        return v

    def is_successful(self):
        return self.success

    def assert_successful(self):
        assert self.success, f"Text is biased - score: {self.score}"

    @property
    def __name__(self):
        return "Bias Score"


def assert_unbiased(
    text: str, minimum_score: float = 0.5
):  # see paper for rationale https://arxiv.org/pdf/2208.05777.pdf
    metric = UnBiasedMetric(minimum_score=minimum_score)
    test_case = LLMTestCase(output=text)
    assert_test(test_case, [metric])



import random
from .metric import Metric
from ..test_case import LLMTestCase
from ..singleton import Singleton


class RandomMetric(Metric, metaclass=Singleton):
    def __init__(self, minimum_score: float = 0.3):
        self.minimum_score = minimum_score

    def measure(self, test_case: LLMTestCase):
        self.score = random.random()
        self.success = self.score >= self.minimum_score
        return self.score

    def is_successful(self):
        return self.success

    @property
    def __name__(self):
        return "Random"
